"""Templatetags for rendering webpack bundle script tags"""  # noqa: INP001

from collections.abc import Iterator
from os import path
from typing import Any

from django import template
from django.conf import settings
from django.http import HttpRequest
from django.utils.safestring import SafeText, mark_safe
from mitol.common.utils.webpack import webpack_public_path
from webpack_loader.utils import get_loader

register = template.Library()


def _get_bundle(request: HttpRequest, bundle_name: str) -> Iterator[dict]:
    """
    Update bundle URLs to handle webpack hot reloading correctly if DEBUG=True

    Args:
        request (django.http.request.HttpRequest): A request
        bundle_name (str): The name of the webpack bundle

    Returns:
        iterable of dict:
            The chunks of the bundle. Usually there's only one but I suppose you could have
            CSS and JS chunks for one bundle for example
    """  # noqa: E501
    if not settings.WEBPACK_DISABLE_LOADER_STATS:
        for chunk in get_loader("DEFAULT").get_bundle(bundle_name):
            chunk_copy = dict(chunk)
            chunk_copy["url"] = path.join(webpack_public_path(request), chunk["name"])  # noqa: PTH118
            yield chunk_copy


@register.simple_tag(takes_context=True)
def render_bundle(
    context: dict[str, Any],
    bundle_name: str,
    added_attrs: str = "",
) -> SafeText:
    """
    Render the script tags for a Webpack bundle

    We use this instead of webpack_loader.templatetags.webpack_loader.render_bundle because we want to substitute
    a dynamic URL for webpack dev environments. Maybe in the future we should refactor to use publicPath
    instead for this.

    Args:
        context (dict): The context for rendering the template (includes request)
        bundle_name (str): The name of the bundle to render
        added_attrs (str): Optional string of HTML attributes to add to the script/link tag

    Returns:
        django.utils.safestring.SafeText:
    """  # noqa: E501
    try:
        bundle = _get_bundle(context["request"], bundle_name)
        return _render_tags(bundle, added_attrs)
    except OSError:
        # webpack-stats.json doesn't exist
        return mark_safe("")


def _render_tags(bundle: Iterator[dict], added_attrs: str = "") -> SafeText:
    """
    Outputs tags for template rendering.
    Adapted from webpack_loader.utils.get_as_tags and webpack_loader.templatetags.webpack_loader.

    Args:
        bundle (iterable of dict): The information about a webpack bundle
        added_attrs (str): Optional string of HTML attributes to add to the script/link tag

    Returns:
        django.utils.safestring.SafeText: HTML for rendering bundles
    """  # noqa: E501, D401

    tags = []
    for chunk in bundle:
        if chunk["name"].endswith((".js", ".js.gz")):
            tags.append(
                ('<script type="text/javascript" src="{}" {} ></script>').format(
                    chunk["url"], added_attrs
                )
            )
        elif chunk["name"].endswith((".css", ".css.gz")):
            tags.append(
                ('<link type="text/css" href="{}" rel="stylesheet" {} />').format(
                    chunk["url"], added_attrs
                )
            )
    return mark_safe("\n".join(tags))  # noqa: S308



from celery import Celery
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string


def get_celery_app() -> Celery:
    """Get the celery app for the project we're running in"""
    app = import_string(settings.MITOL_CELERY_APP_INSTANCE_PATH)

    if not isinstance(app, Celery):
        msg = (
            "Setting MITOL_CELERY_APP_INSTANCE_PATH does not reference "
            "an instance of `celery.Celery`"
        )
        raise ImproperlyConfigured(msg)

    return app



"""Model Serialization utils"""

import json

from django.core.serializers import serialize


def serialize_model_object(obj):  # noqa: RET503
    """
    Serialize model into a dict representable as JSON
    Args:
        obj (django.db.models.Model): An instantiated Django model
    Returns:
        dict:
            A representation of the model
    """
    # serialize works on iterables so we need to wrap object in a list, then unwrap it
    if obj:
        data = json.loads(serialize("json", [obj]))[0]
        serialized = data["fields"]
        serialized["id"] = data["pk"]
        return serialized



import functools
import random
from collections.abc import Callable
from functools import wraps

from django.utils.cache import get_max_age, patch_cache_control
from django_redis import get_redis_connection
from typing_extensions import ParamSpec

P = ParamSpec("P")


def cache_control_max_age_jitter(*args, **kwargs):  # noqa: ARG001
    def _cache_controller(viewfunc):
        @wraps(viewfunc)
        def _cache_controlled(request, *args, **kw):
            # Ensure argument looks like a request.
            if not hasattr(request, "META"):
                raise TypeError(  # noqa: TRY003
                    "cache_control_max_age_jitter didn't receive an HttpRequest. If you are "  # noqa: EM101, E501
                    "decorating a classmethod, be sure to use "
                    "@method_decorator."
                )
            response = viewfunc(request, *args, **kw)
            max_age = get_max_age(response)
            if not max_age:
                max_age = kwargs["max_age"]
            # add random delay upto 5 minutes
            kwargs["max_age"] = max_age + random.randint(1, 600)  # noqa: S311
            patch_cache_control(response, **kwargs)
            return response

        return _cache_controlled

    return _cache_controller


def single_task(
    timeout: int,
    raise_block: bool | None = True,  # noqa: FBT002, FBT001
    key: (str or Callable[[str, P.args, P.kwargs], str]) | None = None,
    cache_name: str | None = "redis",
) -> Callable:
    """
    Only allow one instance of a celery task to run concurrently
    Based on https://bit.ly/2RO2aav

    Args:
        timeout(int): Time in seconds to wait before relinquishing a lock
        raise_block(bool): If true, raise a BlockingIOError when locked
        key(str | Callable): Custom lock name or function to generate one
        cache_name(str): The name of the celery redis cache (default is "redis")

    Returns:
        Callable: wrapped function (typically a celery task)
    """

    def task_run(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            has_lock = False
            client = get_redis_connection(cache_name)
            if isinstance(key, str):
                lock_id = key
            elif callable(key):
                lock_id = key(func.__name__, args, kwargs)
            else:
                lock_id = func.__name__
            lock = client.lock(f"task-lock:{lock_id}", timeout=timeout)
            try:
                has_lock = lock.acquire(blocking=False)
                if has_lock:
                    return_value = func(*args, **kwargs)
                else:
                    if raise_block:
                        raise BlockingIOError
                    return_value = None
            finally:
                if has_lock and lock.locked():
                    lock.release()
            return return_value

        return wrapper

    return task_run



"""Collections utilities"""

from collections.abc import Iterable
from itertools import groupby, islice, tee


def dict_without_keys(d, *omitkeys):
    """
    Returns a copy of a dict without the specified keys

    Args:
        d (dict): A dict that to omit keys from
        *omitkeys: Variable length list of keys to omit

    Returns:
        dict: A dict with omitted keys
    """  # noqa: D401
    return {key: d[key] for key in d.keys() if key not in omitkeys}  # noqa: SIM118


def filter_dict_by_key_set(dict_to_filter, key_set):
    """Takes a dictionary and returns a copy with only the keys that exist in the given set"""  # noqa: E501, D401
    return {key: dict_to_filter[key] for key in dict_to_filter.keys() if key in key_set}  # noqa: SIM118


def first_matching_item(iterable, predicate):
    """
    Gets the first item in an iterable that matches a predicate (or None if nothing matches)

    Returns:
        Matching item or None
    """  # noqa: E501, D401
    return next(filter(predicate, iterable), None)


def find_object_with_matching_attr(iterable, attr_name, value):
    """
    Finds the first item in an iterable that has an attribute with the given name and value. Returns
    None otherwise.

    Returns:
        Matching item or None
    """  # noqa: E501, D401
    for item in iterable:
        try:
            if getattr(item, attr_name) == value:
                return item
        except AttributeError:  # noqa: PERF203
            pass
    return None


def has_equal_properties(obj, property_dict):
    """
    Returns True if the given object has the properties indicated by the keys of the given dict, and the values
    of those properties match the values of the dict
    """  # noqa: E501, D401
    for field, value in property_dict.items():
        try:
            if getattr(obj, field) != value:
                return False
        except AttributeError:  # noqa: PERF203
            return False
    return True


def first_or_none(iterable):
    """
    Returns the first item in an iterable, or None if the iterable is empty

    Args:
        iterable (iterable): Some iterable
    Returns:
        first item or None
    """  # noqa: D401
    return next((x for x in iterable), None)


def max_or_none(iterable):
    """
    Returns the max of some iterable, or None if the iterable has no items

    Args:
        iterable (iterable): Some iterable
    Returns:
        max item or None
    """  # noqa: D401
    try:
        return max(iterable)
    except ValueError:
        return None


def partition(items, predicate=bool):
    """
    Partitions an iterable into two different iterables - the first does not match the given condition, and the second
    does match the given condition.

    Args:
        items (iterable): An iterable of items to partition
        predicate (function): A function that takes each item and returns True or False
    Returns:
        tuple of iterables: An iterable of non-matching items, paired with an iterable of matching items
    """  # noqa: E501
    a, b = tee((predicate(item), item) for item in items)
    return (item for pred, item in a if not pred), (item for pred, item in b if pred)


def partition_to_lists(items, predicate=bool):
    """
    Partitions an iterable into two different lists - the first does not match the given condition, and the second
    does match the given condition.

    Args:
        items (iterable): An iterable of items to partition
        predicate (function): A function that takes each item and returns True or False
    Returns:
        tuple of lists: A list of non-matching items, paired with a list of matching items
    """  # noqa: E501
    a, b = partition(items, predicate=predicate)
    return list(a), list(b)


def unique(iterable):
    """
    Returns a generator containing all unique items in an iterable

    Args:
        iterable (iterable): An iterable of any hashable items
    Returns:
        generator: Unique items in the given iterable
    """  # noqa: D401
    seen = set()
    return (x for x in iterable if x not in seen and not seen.add(x))


def unique_ignore_case(strings):
    """
    Returns a generator containing all unique strings (coerced to lowercase) in a given iterable

    Args:
        strings (iterable of str): An iterable of strings
    Returns:
        generator: Unique lowercase strings in the given iterable
    """  # noqa: E501, D401
    seen = set()
    return (s for s in map(str.lower, strings) if s not in seen and not seen.add(s))


def item_at_index_or_none(indexable, index):
    """
    Returns the item at a certain index, or None if that index doesn't exist

    Args:
        indexable (list or tuple):
        index (int): The index in the list or tuple

    Returns:
        The item at the given index, or None
    """  # noqa: D401
    try:
        return indexable[index]
    except IndexError:
        return None


def item_at_index_or_blank(indexable, index):
    """
    Returns the item at a certain index, or a blank string if that index doesn't exist

    Args:
        indexable (List[str]): A list of strings
        index (int): The index in the list or tuple

    Returns:
        str: The item at the given index, or a blank string
    """  # noqa: D401
    return item_at_index_or_none(indexable, index) or ""


def all_equal(*args):
    """
    Returns True if all of the provided args are equal to each other

    Args:
        *args (hashable): Arguments of any hashable type

    Returns:
        bool: True if all of the provided args are equal, or if the args are empty
    """  # noqa: D401
    return len(set(args)) <= 1


def all_unique(iterable):
    """
    Returns True if all of the provided args are equal to each other

    Args:
        iterable: An iterable of hashable items

    Returns:
        bool: True if all of the provided args are equal
    """  # noqa: D401
    return len(set(iterable)) == len(iterable)


def has_all_keys(dict_to_scan, keys):
    """
    Returns True if the given dict has all of the given keys

    Args:
        dict_to_scan (dict):
        keys (iterable of str): Iterable of keys to check for

    Returns:
        bool: True if the given dict has all of the given keys
    """  # noqa: D401
    return all(key in dict_to_scan for key in keys)


def group_into_dict(items, key_fn):
    """
    Groups items into a dictionary based on a key generated by a given function

    Examples:
        items = [
            Car(make="Honda", model="Civic"),
            Car(make="Honda", model="Accord"),
            Car(make="Ford", model="F150"),
            Car(make="Ford", model="Focus"),
        ]
        group_into_dict(items, lambda car: car.make) == {
            "Honda": [Car(make="Honda", model="Civic"), Car(make="Honda", model="Accord")],
            "Ford": [Car(make="Ford", model="F150"), Car(make="Ford", model="Focus")],
        }

    Args:
        items (Iterable[T]): An iterable of objects to group into a dictionary
        key_fn (Callable[[T], Any]): A function that will take an individual item and produce a dict key

    Returns:
        Dict[Any, T]: A dictionary with keys produced by the key function paired with a list of all the given
            items that produced that key.
    """  # noqa: E501, D401
    sorted_items = sorted(items, key=key_fn)
    return {
        key: list(values_iter) for key, values_iter in groupby(sorted_items, key=key_fn)
    }


def chunks(iterable: Iterable, *, chunk_size: int = 20):
    """
    Yields chunks of an iterable as sub lists each of max size chunk_size.

    Args:
        iterable (iterable): iterable of elements to chunk
        chunk_size (int): Max size of each sublist

    Yields:
        list: List containing a slice of list_to_chunk
    """  # noqa: D401
    chunk_size = max(1, chunk_size)
    iterable = iter(iterable)
    chunk = list(islice(iterable, chunk_size))

    while len(chunk) > 0:
        yield chunk
        chunk = list(islice(iterable, chunk_size))


def matching_item_index(iterable, value_to_match):
    """
    Returns the index of the given value in the iterable

    Args:
        iterable (Iterable): The iterable to search
        value_to_match (Any): The value to match

    Returns:
        int: The index of the matching value

    Raises:
        StopIteration: Raised if the value is not found in the iterable
    """  # noqa: D401
    return next(i for i, value in enumerate(iterable) if value == value_to_match)


def replace_null_values(iterable, replacement):
    """
    Replace any nulls in list or dict values

    Args:
        iterable(Iterable): the iterable to replace any null values
        replacement(any): What to replace invalid values with

    Returns:
        iterable: iterable with null values replaced
    """

    if isinstance(iterable, dict):
        return {
            key: value if value is not None else replacement
            for key, value in iterable.items()
        }
    else:
        return [val if val is not None else replacement for val in iterable]



"""Common default factories"""

from django.contrib.auth import get_user_model
from factory import Faker, SelfAttribute, Trait
from factory.django import DjangoModelFactory
from factory.fuzzy import FuzzyText


class UserFactory(DjangoModelFactory):
    """
    User factory for default django model

    This expects the user model to be django.contrib.auth.models.User or a subclass of it
    """  # noqa: E501

    username = SelfAttribute("email")
    email = FuzzyText(length=20, suffix="@example.com")
    first_name = Faker("first_name")
    last_name = Faker("first_name")

    class Meta:
        model = get_user_model()

    class Params:
        with_global_id = Trait(global_id=Faker("uuid4"))
        with_scim = Trait(
            with_global_id=True,
            scim_external_id=SelfAttribute("global_id"),
            scim_username=SelfAttribute("email"),
        )


class SsoUserFactory(UserFactory):
    """Factory for Users with a global ID."""

    with_global_id = True


class ScimUserFactory(UserFactory):
    """Factory for Users with a global ID + SCIM."""

    with_scim = True



from django.conf import settings


def get_missing_settings(required_settings: list[str]) -> list[str]:
    """Return a list of settings that are missing"""
    return [
        variable
        for variable in required_settings
        if getattr(settings, variable, None) is None
    ]



"""
Common model classes
"""

import copy
from collections.abc import Iterable
from typing import TypeVar, Union

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import transaction
from django.db.models import (
    PROTECT,
    CharField,
    DateTimeField,
    ForeignKey,
    JSONField,
    Model,
    prefetch_related_objects,
)
from django.db.models.query import QuerySet

from mitol.common.utils.datetime import now_in_utc


class TimestampedModelQuerySet(QuerySet):
    """
    Subclassed QuerySet for TimestampedModel
    """

    def update(self, **kwargs):
        """
        Automatically update updated_on timestamp when .update(). This is because .update()
        does not go through .save(), thus will not auto_now, because it happens on the
        database level without loading objects into memory.
        """  # noqa: E501
        if "updated_on" not in kwargs:
            kwargs["updated_on"] = now_in_utc()
        return super().update(**kwargs)


class TimestampedModel(Model):
    """Base model for created_on/updated_on timestamp fields"""

    objects = TimestampedModelQuerySet.as_manager()
    created_on = DateTimeField(auto_now_add=True)  # UTC
    updated_on = DateTimeField(auto_now=True)  # UTC

    class Meta:
        abstract = True


class AuditModel(TimestampedModel):
    """An abstract base class for audit models"""

    acting_user = ForeignKey(settings.AUTH_USER_MODEL, null=True, on_delete=PROTECT)
    data_before = JSONField(blank=True, null=True)
    data_after = JSONField(blank=True, null=True)

    class Meta:
        abstract = True

    @classmethod
    def get_related_field_name(cls):
        """
        Returns:
            str: A field name which links the Auditable model to this model
        """  # noqa: D401
        raise NotImplementedError


class AuditableModel(Model):
    """An abstract base class for auditable models"""

    class Meta:
        abstract = True

    def to_dict(self):
        """
        Returns:
            dict:
                A serialized representation of the model object
        """  # noqa: D401
        raise NotImplementedError

    @classmethod
    def objects_for_audit(cls):
        """
        Returns the correct model manager for the auditable model. This defaults to `objects`, but if
        a different manager is needed for any reason (for example, if `objects` is changed to a manager
        that applies some default filters), it can be overridden.

        Returns:
             django.db.models.manager.Manager: The correct model manager for the auditable model
        """  # noqa: E501, D401
        return cls.objects

    @classmethod
    def get_audit_class(cls):
        """
        Returns:
            class of Model:
                A class of a Django model used as the audit table
        """  # noqa: D401
        raise NotImplementedError

    @transaction.atomic
    def save_and_log(self, acting_user, *args, **kwargs):
        """
        Saves the object and creates an audit object.

        Args:
            acting_user (User):
                The user who made the change to the model. May be None if inapplicable.
        """  # noqa: D401
        before_obj = self.objects_for_audit().filter(id=self.id).first()
        self.save(*args, **kwargs)
        self.refresh_from_db()
        before_dict = None
        if before_obj is not None:
            before_dict = before_obj.to_dict()

        audit_kwargs = dict(  # noqa: C408
            acting_user=acting_user, data_before=before_dict, data_after=self.to_dict()
        )
        audit_class = self.get_audit_class()
        audit_kwargs[audit_class.get_related_field_name()] = self
        audit_class.objects.create(**audit_kwargs)


class SingletonModel(Model):
    """Model class for models representing tables that should only have a single record"""  # noqa: E501

    def save(
        self,
        force_insert=False,  # noqa: FBT002
        force_update=False,  # noqa: FBT002
        using=None,
        update_fields=None,
    ):
        if force_insert and self._meta.model.objects.count() > 0:
            raise ValidationError(  # noqa: TRY003
                f"Only one {self.__class__.__name__} object should exist. Update the existing object instead "  # noqa: EM102, E501
                "of creating a new one."
            )
        return super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields,
        )

    class Meta:
        abstract = True


_ModelClass = TypeVar("_ModelClass", bound=Model)
_PrefetchGenericQuerySet = TypeVar(
    "_PrefetchGenericQuerySet", bound="PrefetchGenericQuerySet"
)


def _items_for_class(
    content_type_field: str,
    items: Iterable[_ModelClass],
    model_cls: type[_ModelClass],
) -> Iterable[_ModelClass]:
    """Returns a list of items that matches a class by content_type"""  # noqa: D401
    return [
        item
        for item in items
        if getattr(item, content_type_field).model_class() == model_cls
    ]


class PrefetchGenericQuerySet(QuerySet):
    """QuerySet supporting for prefetching over generic relationships"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._prefetch_generic_related_lookups = {}
        self._prefetch_generic_done = False

    def prefetch_generic_related(
        self,
        content_type_field: str,
        model_lookups: dict[
            Union[list[type[_ModelClass]], type[_ModelClass]], list[str]
        ],
    ) -> _PrefetchGenericQuerySet:
        """
        Configure prefetch_related over generic relations

        Args:
            content_type_field(str): the field name for the ContentType
            model_lookups(dict of (list of class or class, list of str)):
                a mapping of model classes to lookups

        Returns:
            QuerySet: the new queryset with prefetching configured

        """
        qs = self._chain()  # type: ignore  # noqa: PGH003

        for model_classes, lookups in model_lookups.items():
            model_classes = (  # noqa: PLW2901
                model_classes
                if isinstance(model_classes, Iterable)
                else [model_classes]
            )
            for model_cls in model_classes:
                key = (content_type_field, model_cls)

                qs._prefetch_generic_related_lookups[key] = [  # noqa: SLF001
                    *qs._prefetch_generic_related_lookups.get(key, []),  # noqa: SLF001
                    *lookups,
                ]

        return qs

    def _prefetch_generic_related_objects(self):
        """Prefetch related objects on a per-model basis"""
        for (
            (content_type_field, model_cls),
            lookups,
        ) in self._prefetch_generic_related_lookups.items():
            items = _items_for_class(content_type_field, self._result_cache, model_cls)
            prefetch_related_objects(items, *lookups)
        self._prefetch_generic_done = True

    def _fetch_all(self):
        """Called when a query is evaluated"""  # noqa: D401
        # first fetch non-generic data, this avoid N+1 issues on the generic items themselves  # noqa: E501
        super()._fetch_all()

        if self._prefetch_generic_related_lookups and not self._prefetch_generic_done:
            self._prefetch_generic_related_objects()

    def _clone(self):
        """Clone the queryset"""

        c = super()._clone()
        c._prefetch_generic_related_lookups = copy.deepcopy(  # noqa: SLF001
            self._prefetch_generic_related_lookups
        )
        return c


class UserGlobalIdMixin(Model):
    """
    Mixin that adds a standard global_id definition.

    The `global_id` field points to the SSO ID for the user (so, usually the Keycloak
    ID, which is a UUID). We store it as a string in case the SSO source changes.
    We allow a blank value so we can have out-of-band users - we may want a
     Django user that's not connected to an SSO user, for instance.
    """

    global_id = CharField(
        max_length=36,
        blank=True,
        default="",
        help_text="The SSO ID (usually a Keycloak UUID) for the user.",
    )

    class Meta:
        abstract = True



"""Functions reading and parsing environment variables"""

import importlib
import inspect
import json
import os
from ast import literal_eval
from collections.abc import Callable
from functools import wraps
from typing import Any, NamedTuple, Union

from django.core.exceptions import ImproperlyConfigured


class EnvironmentVariableParseException(ImproperlyConfigured):
    """Environment variable was not parsed correctly"""


class EnvVariable(NamedTuple):
    """Representation of an environment variable"""

    name: str
    default: Any
    description: str
    required: bool
    dev_only: bool
    value: Any
    write_app_json: bool


def var_parser(parser_func: Callable):
    """
    Decorator to create a var parser func

    Args:
        parser_func (callable):
            a function that takes one argument which will be the raw value and
            returns a parsed value or raises an error
    """  # noqa: D401

    @wraps(parser_func)
    def wrapper(  # noqa: PLR0913
        self,
        *,
        name: str,
        description: str,
        default: Any = None,
        required: bool = False,
        dev_only: bool = False,
        write_app_json: bool = True,
    ) -> EnvVariable:
        """
        Get an environment variable

        Args:
            name (str): An environment variable name
            default (str): The default value to use if the environment variable doesn't exist.
            description (str): The description of how this variable is used
            required (bool): Whether this variable is required at runtime
            dev_only (bool): Whether this variable is only applicable in dev environments
            write_app_json (bool): Whether this variable is written to app.json

        Raises:
            ValueError:
                If the environment variable args are incorrect

        Returns:
            any:
                The raw environment variable value
        """  # noqa: E501
        configured_envs = self._configured_vars
        environ = self._env

        if name in configured_envs:
            raise ValueError(f"Environment variable '{name}' was used more than once")  # noqa: EM102, TRY003

        value = environ.get(name, default)
        # attempt to parse the value before we store it in configured_envs
        # this ensures that get_any works since we don't store the various parse attempts until one succeeds  # noqa: E501
        value = parser_func(name, value, default)

        configured_envs[name] = EnvVariable(
            name, default, description, required, dev_only, value, write_app_json
        )

        return value

    return wrapper


def parse_bool(name: str, value: str, default: bool) -> bool:  # noqa: FBT001, ARG001
    """
    Attempts to parse a bool

    Arguments:
        value (str or bool):
            the value as either an unparsed string or a bool in case of a default value

    Raises:
        EnvironmentVariableParseException:
            raised if the value wasn't parsable

    Returns:
        bool:
            parsed value
    """  # noqa: D401

    if isinstance(value, bool):
        return value

    parsed_value = value.lower()
    if parsed_value == "true":
        return True
    elif parsed_value == "false":
        return False

    raise EnvironmentVariableParseException(  # noqa: TRY003
        f"Expected value in {name}={value} to be a boolean"  # noqa: EM102
    )


def parse_int(name: str, value: str, default: int) -> int:
    """
    Attempts to parse a int

    Arguments:
        value (str or int):
            the value as either an unparsed string or an int in case of a default value

    Raises:
        EnvironmentVariableParseException:
            raised if the value wasn't parsable

    Returns:
        int:
            parsed value
    """  # noqa: D401

    if isinstance(value, int) or (value is None and default is None):
        return value

    try:
        parsed_value = int(value)
    except ValueError as ex:
        raise EnvironmentVariableParseException(  # noqa: TRY003
            f"Expected value in {name}={value} to be an int"  # noqa: EM102
        ) from ex

    return parsed_value


def parse_str(name: str, value: str, default: str):  # noqa: ARG001
    """
    Parses a str (identity function)

    Arguments:
        value (str):
            the value as either a str

    Returns:
        str:
            parsed value
    """  # noqa: D401
    return value


def parse_list_literal(
    name: str,
    value: Union[str, list[Any]],
    default: list[Any],  # noqa: ARG001
) -> list[str]:
    """
    Parses a comma separated string into a list

    Argumments:
        value (str or List[Any]):
            the value as a string or list of strings

    Returns:
        list[str]:
            the parsed value
    """  # noqa: D401
    if isinstance(value, list):
        return value

    parse_exception = EnvironmentVariableParseException(
        f"Expected value in {name}={value} to be a list literal"
    )

    try:
        parsed_value = literal_eval(value)
    except (ValueError, SyntaxError) as ex:
        raise parse_exception from ex

    if not isinstance(parsed_value, list):
        raise parse_exception

    return parsed_value


def parse_delimited_list(name, value, default):  # noqa: ARG001
    """
    Parses a comma separated string into a list

    Argumments:
        value (str or list[str]):
            the value as a string or list of strings

    Returns:
        list[str]:
            the parsed value
    """  # noqa: D401
    parsed_value = value
    if isinstance(value, str):
        parsed_value = value.split(",")

    return [item.strip(" ") for item in parsed_value]


CRONTAB_KEYS = ["minute", "hour", "day_of_week", "day_of_month", "month_of_year"]


def parse_crontab_kwargs(name, value, default):  # noqa: ARG001
    """
    Parses a crontab string into structured kwargs

    Argumments:
        value (str):
            the crontab string value
        default (str or dict):
            the default crontab as either a string or dict

    Returns:
        dict:
            the parsed crontab kwargs
    """  # noqa: D401

    if isinstance(value, str):
        units = value.split(" ")

        if len(units) != len(CRONTAB_KEYS):
            raise ImproperlyConfigured(  # noqa: TRY003
                f"Crontab value '{value}' is required to have {len(CRONTAB_KEYS)} units"  # noqa: EM102
            )

        if not all(units):
            raise ImproperlyConfigured(  # noqa: TRY003
                f"Crontab value '{value}' does not have values for all units"  # noqa: EM102
            )

        return dict(zip(CRONTAB_KEYS, units))

    # an empty string would be falsy and we want to catch that above so default the value here  # noqa: E501
    value = value or {}

    if isinstance(value, dict):
        invalid_keys = set(value.keys()) - set(CRONTAB_KEYS)
        if invalid_keys:
            raise ImproperlyConfigured(  # noqa: TRY003
                f"Crontab value has invalid keys: {invalid_keys}"  # noqa: EM102
            )

        return value
    else:
        raise ImproperlyConfigured(f"Crontab value is an invalid type: {type(value)}")  # noqa: EM102, TRY003


class EnvParser:
    """Stateful tracker for environment variable parsing"""

    _env: dict[str, str]
    _configured_vars: dict[str, EnvVariable]
    _imported_modules: list[str]

    def __init__(self):
        self.reset()

    def reset(self):
        """Reset the state"""
        self._env = dict(os.environ)
        self._configured_vars = {}
        self._imported_modules = []

    def reload(self):
        """Reloads the environment"""  # noqa: D401
        # reset the internal state since we're reloading everything that created it
        self.reset()

        # reload the modules we've imported
        for mod in [
            # this needs to be dynamic and inlined because settings.py will import this module  # noqa: E501
            importlib.import_module(os.environ.get("DJANGO_SETTINGS_MODULE")),
            *self._imported_modules,
        ]:
            importlib.reload(mod)

    def validate(self):
        """
        Validates the current configuration

        Raises:
            ImproperlyConfigured:
                If any settings are missing
        """  # noqa: D401
        missing_settings = []

        for env_var in self._configured_vars.values():
            if env_var.required and env_var.value in (None, ""):
                missing_settings.append(env_var.name)  # noqa: PERF401

        if missing_settings:
            raise ImproperlyConfigured(
                "The following settings are missing: {}. You need to add these environment variables in .env file.".format(  # noqa: EM103, E501
                    ", ".join(missing_settings)
                )
            )

    def list_environment_vars(self) -> list[EnvVariable]:
        """
        Get the list of EnvVariables

        Returns:
            list of EnvVariable:
                the list of available env vars
        """
        return self._configured_vars.values()

    def get_features(self, prefix: str = "FEATURE_") -> dict[str, bool]:
        """
        Get the list of features enabled for this app

        Args:
            prefix (str):
                feature prefix string

        Returns:
            dict of bool:
                dictionary of feature flags
        """
        return {
            key[len(prefix) :]: self.get_bool(
                name=key,
                default=False,
                description=f"Feature {key}",
                write_app_json=False,
            )
            for key in list(self._env.keys())
            if key.startswith(prefix)
        }

    def init_app_settings(self, *, namespace: str, site_name: str, **kwargs):
        """
        Configure the app static settings

        Args:
            namespace (str):
                the app settings namespace
            site_name (str):
                The name of the site
        """

        # an optional scope can be passed, or this will default to the calling scope's globals  # noqa: E501
        scope = kwargs.pop("scope", inspect.currentframe().f_back.f_globals)

        scope["APP_SETTINGS_NAMESPACE"] = self.get_string(
            name="APP_SETTINGS_NAMESPACE",
            default=namespace,
            description="App environment variable namespace",
            write_app_json=False,
        )
        scope["SITE_NAME"] = self.get_string(
            name="SITE_NAME", default=site_name, description="Site name"
        )

    def get_site_name(self):
        """Return the site name"""
        site_name = self._configured_vars.get("SITE_NAME")
        if not site_name:
            raise ImproperlyConfigured(  # noqa: TRY003
                "Site name isn't set, add a call to init_app_settings()"  # noqa: EM101
            )
        return site_name.value

    def app_namespaced(self, setting_key: str) -> str:
        """
        Return a namespaced setting key

        Example: app_namespaced("KEY") -> "APP_KEY"
        """
        namespace = self._configured_vars.get("APP_SETTINGS_NAMESPACE")
        if not namespace:
            raise ImproperlyConfigured(  # noqa: TRY003
                "App settings namespace isn't set, add a call to init_app_settings()"  # noqa: EM101
            )
        return f"{namespace.value}_{setting_key}"

    def import_settings_modules(self, *module_names: str, **kwargs):
        """
        Import settings from modules

        Usage:
            import_settings_modules("module1.settings", "module2.settings")
            import_settings_modules("module1.settings", "module2.settings", scope=globals())
        """  # noqa: E501

        # an optional scope can be passed, or this will default to the calling scope's globals  # noqa: E501
        scope = kwargs.pop("scope", inspect.currentframe().f_back.f_globals)

        # this function imports modules and then walks each, adding uppercased vars
        # to the calling settings module (typically settings.py in a django project)
        for module_name in module_names:
            mod = importlib.import_module(module_name)

            # track this module to support reloading
            self._imported_modules.append(mod)

            # walk the module and get only the constant-cased variables
            for setting_name in dir(mod):
                if setting_name.isupper():
                    setting_value = getattr(mod, setting_name)
                    scope[setting_name] = setting_value

    get_string = var_parser(parse_str)
    get_bool = var_parser(parse_bool)
    get_int = var_parser(parse_int)
    get_list_literal = var_parser(parse_list_literal)
    get_delimited_list = var_parser(parse_delimited_list)
    get_crontab_kwargs = var_parser(parse_crontab_kwargs)


env = EnvParser()

# methods below are our exported module interface
get_string = env.get_string
get_int = env.get_int
get_bool = env.get_bool
get_list_literal = env.get_list_literal
get_delimited_list = env.get_delimited_list
get_crontab_kwargs = env.get_crontab_kwargs

reload = env.reload
validate = env.validate
list_environment_vars = env.list_environment_vars
get_features = env.get_features
init_app_settings = env.init_app_settings
app_namespaced = env.app_namespaced
get_site_name = env.get_site_name
import_settings_modules = env.import_settings_modules


def generate_app_json():
    """
    Generate a new app.json data structure in-memory using app.base.json and settings.py

    Returns:
        dict:
            object that can be serialized to JSON for app.json
    """
    with open("app.base.json") as app_template_json:  # noqa: PTH123
        config = json.load(app_template_json)

    for env_var in list_environment_vars():
        if env_var.dev_only or not env_var.write_app_json:
            continue

        if env_var.name not in config["env"]:
            config["env"][env_var.name] = {}

        config["env"][env_var.name].update(
            {"description": env_var.description, "required": env_var.required}
        )

    return config



"""common constants"""

USERNAME_SEPARATOR = "-"
# Characters that should be replaced by the specified separator character
USERNAME_SEPARATOR_REPLACE_CHARS = "\\s_"
# Characters that should be removed entirely from the full name to create the username
USERNAME_INVALID_CHAR_PATTERN = (
    rf"[^\w{USERNAME_SEPARATOR_REPLACE_CHARS}{USERNAME_SEPARATOR}]|[\d]"
)

USERNAME_TURKISH_I_CHARS = r"[ıİ]"
USERNAME_TURKISH_I_CHARS_REPLACEMENT = "i"

# Pattern for chars to replace with a single separator. The separator character itself
# is also included in this pattern so repeated separators are squashed down.
USERNAME_SEPARATOR_REPLACE_PATTERN = (
    rf"[{USERNAME_SEPARATOR_REPLACE_CHARS}{USERNAME_SEPARATOR}]+"
)

USERNAME_MAX_LEN = 30
USERNAME_COLLISION_ATTEMPTS = 10



"""User utilities."""

import logging
import re

from django.contrib.auth import get_user_model
from django.db import IntegrityError

from mitol.common.constants import (
    USERNAME_COLLISION_ATTEMPTS,
    USERNAME_INVALID_CHAR_PATTERN,
    USERNAME_MAX_LEN,
    USERNAME_SEPARATOR,
    USERNAME_SEPARATOR_REPLACE_PATTERN,
    USERNAME_TURKISH_I_CHARS,
    USERNAME_TURKISH_I_CHARS_REPLACEMENT,
)

from .helpers import max_or_none

log = logging.getLogger(__name__)


def _reformat_for_username(string):
    """Removes/substitutes characters in a string to make it suitable as a
    username value.

    Args:
        string (str): A string
    Returns:
        str: A version of the string with username-appropriate characters
    """
    cleaned_string = re.sub(USERNAME_INVALID_CHAR_PATTERN, "", string)
    cleaned_string = re.sub(
        USERNAME_TURKISH_I_CHARS,
        USERNAME_TURKISH_I_CHARS_REPLACEMENT,
        cleaned_string,
    )
    return (
        re.sub(USERNAME_SEPARATOR_REPLACE_PATTERN, USERNAME_SEPARATOR, cleaned_string)
        .lower()
        .strip(USERNAME_SEPARATOR)
    )


def is_duplicate_username_error(exc):
    """Return True if the given exception indicates that there was an attempt
    to save a User record with an already-existing username.

    Args:
        exc (Exception): An exception

    Returns:
        bool: Whether or not the exception indicates a duplicate username error
    """
    return re.search(r"\(username\)=\([^\s]+\) already exists", str(exc)) is not None


def _find_available_username(  # noqa: RET503
    initial_username_base,
    model=None,
    username_field="username",
    max_length=USERNAME_MAX_LEN,
):
    """Return a username with the lowest possible suffix given some
    base username. If the applied suffix makes the username longer than the
    username max length, characters are removed from the right of the username
    to make room.

    EXAMPLES:
    initial_username_base = "johndoe"
        Existing usernames = "johndoe"
        Return value = "johndoe1"
    initial_username_base = "johndoe"
        Existing usernames = "johndoe", "johndoe1" through "johndoe5"
        Return value = "johndoe6"
    initial_username_base = "abcdefghijklmnopqrstuvwxyz"  (26 characters,
        assuming 26 character max)
        Existing usernames = "abcdefghijklmnopqrstuvwxyz"
        Return value = "abcdefghijklmnopqrstuvwxy1"  # pragma: allowlist secret
    initial_username_base = "abcdefghijklmnopqrstuvwxy"
        (25 characters long, assuming 26 character max)
        Existing usernames = "abc...y", "abc...y1" through "abc...y9"
        Return value = "abcdefghijklmnopqrstuvwx10"  # pragma: allowlist secret

    Args:
         initial_username_base (str): Base username to start with
         model: The model class to query (defaults to User model)
         username_field (str): The field name to use for username queries
            (defaults to 'username')
         max_length (int): The maximum allowed username length
            (defaults to USERNAME_MAX_LEN)
    Returns:
        str: An available username
    """
    if model is None:
        model = get_user_model()

    # Keeps track of the number of characters that must be cut from the
    # username to be less than the username max length when the
    # suffix is applied.
    letters_to_truncate = 0 if len(initial_username_base) < max_length else 1
    # Any query for suffixed usernames could come up empty. The minimum suffix
    # will be added to the username in that case.
    current_min_suffix = 1
    while letters_to_truncate < len(initial_username_base):
        username_base = initial_username_base[
            0 : len(initial_username_base) - letters_to_truncate
        ]
        # Find usernames that match the username base and have a numerical
        # suffix, then find the max suffix
        filter_kwargs = {f"{username_field}__regex": rf"^{username_base}\d+$"}
        existing_usernames = model.objects.filter(**filter_kwargs).values_list(
            username_field, flat=True
        )

        suffixes = []
        for username in existing_usernames:
            match = re.search(r"\d+$", username)
            if match is not None:
                try:
                    suffixes.append(int(match.group()))
                except (ValueError, AttributeError):
                    continue

        max_suffix = max_or_none(suffixes) if suffixes else None

        if max_suffix is None:
            return "".join([username_base, str(current_min_suffix)])
        else:
            next_suffix = max_suffix + 1
            candidate_username = "".join([username_base, str(next_suffix)])
            # If the next suffix adds a digit and causes the username to
            # exceed the character limit, keep searching.
            if len(candidate_username) <= max_length:
                return candidate_username
        # At this point, we know there are no suffixes left to add to this
        # username base that was tried,
        # so we will need to remove letters from the end of that username base
        # to make room for a longer suffix.
        letters_to_truncate = letters_to_truncate + 1
        available_suffix_digits = max_length - (
            len(initial_username_base) - letters_to_truncate
        )
        # If there is space for 4 digits for the suffix, the minimum value it
        # could be is 1000, or 10^3
        current_min_suffix = 10 ** (available_suffix_digits - 1)


def usernameify(full_name, email="", max_length=USERNAME_MAX_LEN):
    """Public API for username generation Generate a username based on a
    full name, or an email address as a fallback.

    Args:
        full_name (str): A full name (i.e.: User.name)
        email (str): An email address to use as a fallback if the full name
            produces a blank username
        max_length (int): The maximum allowed username length
            (defaults to USERNAME_MAX_LEN)
    Returns:
        str: A generated username
    Raises:
        ValueError: Raised if generated username was blank after trying both
            the full name and email
    """
    username = _reformat_for_username(full_name)

    if not username and email:
        log.error(
            "User's full name could not be used to generate a username "
            "(full name: '%s'). "
            "Trying email instead...",
            full_name,
        )
        username = _reformat_for_username(email.split("@")[0])
    if not username:
        raise ValueError(
            "Username could not be generated (full_name: '{}', email: '{}')".format(  # noqa: EM103, UP032
                full_name, email
            )
        )
    return username[0:max_length]


def create_user_with_generated_username(  # noqa: PLR0913
    serializer,
    initial_username,
    username_field,
    max_length,
    model=None,
    attempts_limit=USERNAME_COLLISION_ATTEMPTS,
):
    """Public API for user creation with auto-generated username
    Creates a User with a given username, and if there is a User that already
    has that username, finds an available username and
    reattempts the User creation.

    Args:
        serializer: A serializer instance that has been instantiated
            with user data and has passed initial validation
        initial_username (str): The initial username to attempt to
            save the User with
        username_field (str): The field name to use for the username
        max_length (int): The maximum allowed username length
        model: The model class to pass to find_username_func (defaults to None)
        attempts_limit (int): Maximum number of attempts to create user

    Returns:
        User or None: The created User (or None if the User could not be
        created in the number of retries allowed)
    """
    created_user = None
    username = initial_username
    attempts = 0

    if len(username) < 2:  # noqa: PLR2004
        username = username + "11"

    while created_user is None and attempts < attempts_limit:
        try:
            created_user = serializer.save(username=username)
        except IntegrityError as exc:  # noqa: PERF203
            if not is_duplicate_username_error(exc):
                raise
            username = _find_available_username(
                initial_username,
                model=model,
                username_field=username_field,
                max_length=max_length,
            )
        finally:
            attempts += 1
    return created_user



"""Celery settings"""

from mitol.common.envs import get_string

MITOL_CELERY_APP_INSTANCE_PATH = get_string(
    name="MITOL_CELERY_APP_INSTANCE_PATH",
    default="main.celery.app",
    description="Path to the celery app instance",
    required=True,
)



"""helper functions for utilities"""


def max_or_none(iterable):
    """Return the max of some iterable, or None if the iterable has no items

    Args:
        iterable (iterable): Some iterable
    Returns:
        max item or None
    """
    try:
        return max(iterable)
    except ValueError:
        return None



"""Base utilities"""

import abc


def base_register_subclasses_factory(*mixin_classes):
    class BaseRegisterSubclasses(abc.ABC, *mixin_classes):
        _SUBCLASSES = {}

        def __init_subclass__(cls, *, subclass_type, **kwargs):
            super().__init_subclass__()
            if subclass_type in cls._SUBCLASSES:
                raise TypeError(f"{subclass_type} has already been defined")  # noqa: EM102, TRY003
            cls._SUBCLASSES[subclass_type] = cls

        @classmethod
        def get_all_subclasses(cls):
            return cls._SUBCLASSES.values()

        @classmethod
        def get_subclass_by_type(cls, subclass_type):
            return cls._SUBCLASSES[subclass_type]

        @classmethod
        def get_all_types(cls):
            return cls._SUBCLASSES.keys()

    return BaseRegisterSubclasses
